import logging

from .aria2 import Aria2
from .session import Session
from config import YAMLConfig, Constants
import os
import re
import subprocess
import threading

from m3u8_dl.parser import Parser

logger = logging.getLogger(__name__)


class Downloader(object):

    def __init__(self):
        self._aria2p = None
        self._m3u8_download_list = None
        self._downloading = False
        self._total_segments = 0

    @property
    def aria2p(self):
        return self._aria2p

    @property
    def m3u8_download_list(self):
        return self._m3u8_download_list

    def download(self, m3u8s, session=True):
        self._m3u8_download_list = Parser.parse(m3u8s)
        if len(m3u8s) == 0:
            logger.info('Download abort: m3u8s is empty and there is nothing to download.')
            return
        logger.info('Get %d download tasks, preparing...' % len(m3u8s))
        if not self.aria2p:
            logger.debug('Aria2c is not running.')
            aria2 = Aria2.get_instance()
            self._aria2p = aria2.start()
        # save session
        if session:
            Session.save(m3u8s[0]['name'])
        self._downloading = True
        self.aria2p.purge_all()
        for l in self._m3u8_download_list:
            download_dir = Downloader._get_download_dir(l['name'])
            logger.debug(f'{l["name"]} dir: {download_dir}')
            uris = l['segments']
            download_details = []
            for u in uris:
                self._total_segments += 1
                download = self.aria2p.add_uris([u], {
                    'dir': download_dir,
                    'user_agent': Constants.user_agent
                })
                download_details.append(download)
            logger.info(f'\'{l["name"]}\' is added to Aria2.')
        logger.info('All videos are added to Aria2, please wait for completion...')
        self._download_monitor()
        self.listen_to_notifications()

    def listen_to_notifications(self):
        def complete(api, gid):
            active_num = self.aria2p.get_stats().num_active
            if active_num == 0:
                logger.debug(f'All downloaded.')
                self._downloading = False

        if self.aria2p.get_stats().num_active == 0:
            logger.info('There is nothing to download, listener will be stopped.')
            self.aria2p.stop_listening()
        else:
            self.aria2p.listen_to_notifications(on_download_complete=complete)
        Aria2.get_instance().shutdown()

    def _download_completed(self):
        # for security, all operations should wait for download monitor closed
        self.aria2p.purge_all()
        self.aria2p.stop_listening()
        self.post_download()

    def _download_monitor(self):
        stats = self.aria2p.get_stats()
        logger.debug('<Aria2p stats> num_active: %s, num_waiting: %s, download_speed: %s' % (
            stats.num_active, stats.num_waiting, stats.download_speed
        ))
        not_completed_num = stats.num_active + stats.num_waiting
        completed_num = self._total_segments - not_completed_num
        completed_percentage = completed_num / self._total_segments * 100
        if stats.download_speed < 1000:
            download_speed = '%.2f B/s' % stats.download_speed
        elif 1000 <= stats.download_speed < 1000000:
            ds = stats.download_speed / 1000
            download_speed = '%.2f KB/s' % ds
        else:
            ds = stats.download_speed / 1000000
            download_speed = '%.2f MB/s' % ds
        logger.info('Complete: %.2f%% (%d/%d), download speed: %s' %
                    (completed_percentage, completed_num, self._total_segments, download_speed))
        if self._downloading:
            threading.Timer(1.0, self._download_monitor).start()
        else:
            self._download_completed()

    def post_download(self):
        logger.info('performing post download...')
        # check ffmpeg installation status
        outputs = subprocess.check_output('ffmpeg -version', shell=True).decode(Constants.coding)
        if re.search('ffmpeg version', outputs) is None:
            raise Exception('FFmpeg is not installed.')
        # merge *.tz to mp4
        for l in self.m3u8_download_list:
            download_dir = Downloader._get_download_dir(l['name'])
            filename_of_segments = []
            # uri to ts filename, e.g. https://xxx/out000.ts -> out000.ts
            for s in l['segments']:
                filename_of_segments.append(Downloader._uri_to_filename(s))
            # generate merge info
            merge_file = Downloader._generate_ts_merge_info(download_dir, filename_of_segments)
            # merge
            if merge_file is not None:
                Downloader.merge_ts(merge_file, YAMLConfig.get()["aria2"]["download_path"] + os.sep + l["name"] + ".ts")
            # remove temp directory
            subprocess.call(f'rm -rf "{download_dir}"', shell=True)

    @staticmethod
    def _get_download_dir(name):
        return '%s%s%s' % (YAMLConfig.get()['aria2']['download_path'], os.sep, name)

    @staticmethod
    def _uri_to_filename(uri):
        filename = re.search(r'(?<=/hls/\d{4}/\d{2}/\d{2}/\w{8}/).*\.ts$', uri)
        if filename is None:
            raise ValueError(f'{uri} cannot be parsed to ts file')
        filename = filename.group(0)
        logger.debug(f'{uri} is parsed to {filename}')
        return filename

    @staticmethod
    def _generate_ts_merge_info(download_dir, segments):
        merge_file_path = '%s%s%s' % (download_dir, os.sep, 'merge.txt')
        try:
            with open(merge_file_path, 'w') as f:
                for s in segments:
                    f.write(f'file \'{s}\'\n')
            logger.debug(f'merge info has been created at {merge_file_path}')
        except FileNotFoundError:
            logger.error('Merge info not exists: %s' % merge_file_path)
            return None
        return merge_file_path

    @staticmethod
    def merge_ts(merge_file, output):
        cmd = f'ffmpeg -f concat -i "{merge_file}" -c copy "{output}"'
        process = subprocess.Popen(cmd, shell=True,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        out, err = process.communicate()
        errcode = process.returncode
        logger.debug(f'errcode: {errcode}')
        if errcode == 0:
            logger.info(f'Video is merged successfully and saved into: {output}')
        else:
            logger.error(f'FFpmeg error: \n{err.decode(Constants.coding)}')



from collections import OrderedDict

import requests
import logging
from config import Constants, YAMLConfig

import re

from m3u8_dl import Session

logger = logging.getLogger(__name__)


class Collector:

    boundary_str = 'PvgqibigcrE8PgXs'

    headers = OrderedDict(
        (
            ("Host", None),
            ("Connection", "keep-alive"),
            ("Upgrade-Insecure-Requests", "1"),
            ("User-Agent", Constants.user_agent),
            (
                "Accept",
                "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;"
                "q=0.8,application/signed-exchange;v=b3",
            ),
            ("Accept-Encoding", "gzip, deflate"),
            ("Accept-Language", "de-DE,de;q=0.9,en-US;q=0.8,en;q=0.7"),

        )
    )

    @staticmethod
    def get_name_and_details_url(html):
        html = re.sub(r'(?<=<a href="html_data/\d{3}/\d{4}/\d{7}\.html") id="\w{14}">', Collector.boundary_str, html)
        logger.debug(f'reasy html with boundary str: \n{html}')
        details_url = re.findall(r'(?<=<a href=")html_data.*?(?="%s)' % Collector.boundary_str, html)
        name = re.findall(r'(?<=%s).*?(?=</a>)' % Collector.boundary_str, html)
        if len(details_url) != len(name):
            logger.error(f'name arr: {name}')
            logger.error(f'details_url: {details_url}')
            raise Exception('regular process for name and details_url failed: the length of them is not equal.')
        name_and_details_url_arr = []
        for i in range(len(details_url)):
            name_and_details_url_arr.append({
                'name': name[i],
                'details_url': '%s%s%s' % (Constants.bt7086_url, '/', details_url[i])
            })
        logger.debug(f'name_and_details_url_arr: {name_and_details_url_arr}')
        return name_and_details_url_arr

    @staticmethod
    def get_m3u8(url):
        logger.info('Reasy is getting video id from \'%s\'' % url)
        response = requests.get(url=url, headers=Collector.headers)
        if response.status_code != 200:
            raise Exception('BT7086 is unreachable, please check out your network.')
        html = response.content.decode(Constants.coding)
        # logger.debug(f'reasy details html: \n{html}')
        m3u8_id = re.search(r'(?<=/\?id=).*?(?=")', html)
        if m3u8_id is None:
            raise Exception(f'm3u8 id cannot be found in {url}')
        m3u8_id = m3u8_id.group(0)
        logger.debug(f'm3u8 id({m3u8_id}) is get from {url}')
        return m3u8_id

    @staticmethod
    def get_m3u8s(session=True, maximum_download_num=30):
        logger.info('Getting m3u8s from bt7086...')
        response = requests.get(url=Constants.bt7086_url + '/thread.php?fid=111', headers=Collector.headers)
        if response.status_code != 200:
            raise Exception('BT7086 is unreachable, please check out your network.')
        html = response.content.decode(Constants.coding)
        logger.debug(f'BT7086 html: \n{html}')
        name_and_details_url_arr = Collector.get_name_and_details_url(html)
        m3u8s = []
        if len(name_and_details_url_arr) > maximum_download_num:
            logger.info(f'The length of current list is greater than maximum_download_num')
            name_and_details_url_arr = name_and_details_url_arr[:maximum_download_num]
        last_download_item = None
        if session:
            last_download_item = Session.read()
        for i in name_and_details_url_arr:
            if session and last_download_item is not None and last_download_item == i["name"]:
                logger.info('Detected repeated download item, all download tasks after \'%s\' will be cancelled.'
                            % last_download_item)
                break
            logger.debug(f'name: {i["name"]}\ndetails_url: {i["details_url"]}')
            m3u8_id = Collector.get_m3u8(i["details_url"])
            m3u8_url = 'https://m3u8.cdnpan.com/%s.m3u8' % m3u8_id
            m3u8s.append({
                'name': i["name"],
                'url': m3u8_url
            })
        logger.debug(f'm3u8s: {m3u8s}')
        return m3u8s



import click
import logging
from m3u8_dl.downloader import Downloader
from collector import Collector
from version import __version__ as ver
import logging

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(None, '-v', '--version',
                      message='reasy version %s' % ver)
def cli():
    pass


def verbose_callback(ctx, param, value):
    if value:
        click.echo('Enables verbose.')
        logging.basicConfig(
            format='%(asctime)s [%(levelname)s] %(message)s (%(filename)s:%(lineno)s)',
            datefmt='%Y-%m-%d %H:%M:%S',
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format='%(asctime)s [%(levelname)s] %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
            level=logging.INFO,
        )


@cli.command()
@click.option('--no-session',
              is_flag=True,
              help='Disable session, which may cause repeated download.')
@click.option('--maximum-download-num', '-m',
              default=30,
              type=int)
@click.option('--verbose',
              is_flag=True,
              expose_value=False,
              is_eager=True,
              callback=verbose_callback)
def run(no_session, maximum_download_num):
    session = not no_session
    if no_session:
        click.echo('Session has been disabled.')
    dl = Downloader()
    dl.download(Collector.get_m3u8s(session, maximum_download_num), session=session)



from setuptools import setup, find_packages
from os.path import dirname, abspath, join
from version import __version__ as version

install_reqs = [req for req in open(abspath(join(dirname(__file__), 'requirements.txt')))]

setup(
    name='reasy',
    version=version,
    author='dekofejld',
    author_email='dekofejld@gmail.com',
    py_modules=['reasy', 'version', 'm3u8-dl', 'config', 'collector'],
    packages=find_packages(),
    license='MIT',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    install_requires=install_reqs,
    description="A tool for quick downloading videos from BT7086",
    entry_points='''
        [console_scripts]
        reasy=reasy:cli
    ''',
)
